import yaml
import sqlite3
import re


def read_yaml_file(file_path):
    with open(file_path, "r") as f:
        return yaml.safe_load(f)


def get_database_config(databases, db_name):
    for db in databases:
        if db["name"] == db_name:
            return db
    raise ValueError(
        f"[ERROR] Database config for '{db_name}' not found in database.yaml."
    )


def execute_sqlite_query(db_path, query):
    try:
        conn = sqlite3.connect(db_path)
        cursor = conn.cursor()
        cursor.execute(query)
        result = cursor.fetchone()
        conn.close()
        return float(result[0]) if result and result[0] is not None else None
    except Exception as e:
        print(f"[ERROR] SQLite query failed on {db_path}: {e}")
        return None


def is_safe_query(query):
    # Remove leading spaces and brackets
    cleaned_query = query.strip().lower()
    # Match only queries that start with "select"
    return re.match(r"^\(*\s*select", cleaned_query) is not None


def process_database_queries(queries_file, databases_file):
    # Get queries from queries file
    queries_config = read_yaml_file(queries_file)
    # Get database from database file
    databases_config = read_yaml_file(databases_file)

    commands = queries_config.get("commands", [])
    databases = databases_config.get("databases", [])

    all_results = []

    for cmd in commands:
        try:
            db_conf = get_database_config(databases, cmd["database"])

            if db_conf["db_type"] != "sqlite":
                print(
                    f"[WARN] Unsupported database type: {db_conf['db_type']} in command {cmd['name']}"
                )
                continue

            db_path = db_conf["db_connection"]
            query = cmd["query"]
            labels = cmd.get('labels', [])
            label_values = cmd.get('label_values', [])
            main_label = cmd.get('main_label', 'default_db_metric')

            # check if query is safe
            if is_safe_query(query):
                value = execute_sqlite_query(db_path, query)
            else:
                print(f"[WARN] Unsupported query type: {query}")
                return None

            if not isinstance(label_values, list):
                label_values = [label_values]

            result = {
                'result': [{
                    'label': label_values,
                    'value': value
                }],
                'labels': labels,
                'main_label': main_label
            }
            all_results.append(result)
        except Exception as e:
            print(
                f"[ERROR] Processing command '{cmd.get('name', 'unknown')}' failed: {e}"
            )
    return all_results



import os
import ast
import importlib.util
from typing import Dict, List, Any


def get_python_files(folder_path: str) -> List[str]:

    return [ os.path.join(root,file) for root, _ , files  in os.walk(folder_path) for file in files if file.endswith('.py') ]

def extract_functions_with_code(file_path: str) -> Dict[str, str]:
    with open(file_path, "r") as file:
        file_content = file.read()  # Read file content once

    tree = ast.parse(file_content)

    functions = {}
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            function_name = node.name
            function_code = ast.get_source_segment(file_content, node)
            functions[function_name] = function_code
    return functions

def analyze_dependencies(function_code: str) -> List[str]:
    try:
        tree = ast.parse(function_code)
    except SyntaxError as e:
        print(f"Syntax error in function code: {e}")
        return []

    dependencies = []
    for node in ast.walk(tree):
        if isinstance(node, ast.Call):
            if isinstance(node.func, ast.Name):
                dependencies.append(node.func.id)
            elif isinstance(node.func, ast.Attribute):
                dependencies.append(node.func.attr)

    return dependencies

def find_function_dependencies(file_path: str) -> Dict[str, List[str]]:
    functions_with_code = extract_functions_with_code(file_path)
    function_dependencies = {}

    for func_name, func_code in functions_with_code.items():
        dependencies = analyze_dependencies(func_code)
        function_dependencies[func_name] = dependencies

    return function_dependencies

def topological_sort(dependency_graph: Dict[str, List[str]]) -> List[str]:
    visited = set()
    stack = []

    def visit(node):
        if node not in visited:
            visited.add(node)
            for neighbor in dependency_graph.get(node, []):
                visit(neighbor)
            stack.append(node)

    for node in dependency_graph:
        visit(node)

    return stack[::-1]

def execute_function(file_path: str, function_name: str, attributes: Dict[str,Any]):
    dependencies = find_function_dependencies(file_path)
    sorted_functions = topological_sort(dependencies)

    if function_name not in sorted_functions:
        return

    module_name = os.path.splitext(os.path.basename(file_path))[0]
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)   

    for func in sorted_functions:
        if hasattr(module, func):
            func_to_call = getattr(module, func)
            if callable(func_to_call):
                try:
                    args = attributes.get(func,())
                    if not isinstance(args,tuple):
                        args = (args,)
                    result = func_to_call(*args)  
                    if func == function_name:
                        return result
                except Exception as e:
                    print(f"Error executing {func}: {e}")

def main(folder_path: str, function_name: str, attributes: Dict[str,Any]):
    python_files = get_python_files(folder_path)

    for file_path in python_files:
        return execute_function(file_path, function_name,attributes)


def read_function_file(custom_path):    
    with open(custom_path,'r') as file:
        lines = file.readlines()
    return lines

def get_env_dir():
    current_path = os.getcwd()
    return os.path.abspath(os.path.join(current_path,'env/bin/activate'))

def install_requirements(requirements_path):
    env_dir = get_env_dir()
    command = f"source {env_dir}; pip3 install -r {requirements_path}"
    os.system(command)

def load_variable_from_file(file_path, variable_name="ARGS"):
    spec = importlib.util.spec_from_file_location("module_name", file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return getattr(module, variable_name, None)

def execute_functions(folder_path,file_path):
    functions = read_function_file(file_path)
    args= { 'collect_disk_metrics' : ()}
    result = [ main(folder_path, function_name.strip('\n'), args) for function_name in functions ]
    return result


import subprocess
import yaml
import re
from flexmetric.logging_module.logger import get_logger

logger = get_logger(__name__)

logger.info("prometheus is running") 

def read_commands_from_yaml(file_path):
    with open(file_path, 'r') as f:
        config = yaml.safe_load(f)
    return config.get('commands', [])

def execute_command_with_timeout(command, timeout):
    try:
        result = subprocess.run(command, shell=True, capture_output=True, text=True, timeout=timeout)
        if result.returncode != 0:
            logger.info(f"Exception in running the command {command}")
            return ''
        return result.stdout.strip()
    except subprocess.TimeoutExpired:
        return ''
    except Exception as ex:
        logger.error(f"Exception : {ex}")
        return ''

def parse_command_output(raw_output, label_column, value_column, fixed_label_value):
    result_list = []
    lines = raw_output.strip().splitlines()

    for line in lines:
        parts = line.strip().split()
        if not parts:
            continue

        if label_column == 'fixed':
            label = fixed_label_value or 'label'
        else:
            try:
                label = parts[label_column]
            except IndexError:
                label = 'unknown'

        try:
            raw_value = parts[value_column]
            cleaned_value = re.sub(r'[^\d\.\-]', '', raw_value)
            value = float(cleaned_value) if cleaned_value else 1
        except (IndexError, ValueError):
            value = 1

        result_list.append({'label': label, 'value': value})

    return result_list

def process_single_command(cmd_info):
    command = cmd_info['command']
    timeout = cmd_info.get('timeout_seconds', 30)
    labels = cmd_info.get('labels', [])
    label_columns = cmd_info.get('label_columns', [])
    value_column = cmd_info.get('value_column', 0)
    main_label = cmd_info.get('main_label', 'default_metric')

    raw_output = execute_command_with_timeout(command, timeout)
    if not raw_output:
        logger.warning(f"No results for command {command}")
        return None

    lines = raw_output.strip().split('\n')
    if not lines:
        logger.error(f"No valid lines returned from command: {command}")
        return None
    result_list = []
    for line in lines:
        parts = line.split()
        try:
            label_columns_value = []
            for value in label_columns:
                label_columns_value.append(parts[value])
            result_list.append({ "label": label_columns_value, "value": parts[value_column] })
        except Exception as e:
            logger.error(f"Error parsing line: '{line}' → {e}")
            continue
    return {
        'result': result_list,
        'labels': labels,
        'main_label': main_label
    }



def is_command_safe(command):
    blacklist = ['rm', 'reboot', 'shutdown', 'halt', 'poweroff', 'mkfs', 'dd']
    for dangerous_cmd in blacklist:
        if dangerous_cmd in command.split():
            return False
    return True

def process_commands(config_file):
    commands = read_commands_from_yaml(config_file)
    all_results = []

    for cmd_info in commands:
        command = cmd_info.get('command', '')
        if not command:
            logger.error("Command is missing in the configuration.")
            continue

        if not is_command_safe(command):
            logger.warning(f"Command '{command}' is not allowed and will not be executed.")
            continue

        try:
            formatted_result = process_single_command(cmd_info)
            if formatted_result:
                all_results.append(formatted_result)
        except KeyError as e:
            logger.error(f"Missing key in command configuration: {e}. Command: {cmd_info}")
        except Exception as e:
            logger.error(f"An error occurred while processing command '{command}': {e}")

    return all_results

# # Example usage:
# if __name__ == "__main__":
#     results = process_commands('/Users/nlingadh/code/custom_prometheus_agent/src/commands.yaml')
#     print(results)



from collections import deque
import threading
import time
import sched

class ExpiringMetricQueue:
    def __init__(self, expiry_seconds=60, cleanup_interval=5):
        self.queue = deque()
        self.expiry_seconds = expiry_seconds
        self.cleanup_interval = cleanup_interval
        self.lock = threading.Lock()
        self.scheduler = sched.scheduler(time.time, time.sleep)
        self._start_cleanup()
    
    def put(self, metric):
        metric['timestamp'] = time.time()
        with self.lock:
            self.queue.append(metric)
    
    def pop_all(self):
        with self.lock:
            items = list(self.queue)
            self.queue.clear()
            cleaned_items = []
            for item in items:
                cleaned_item = {k: v for k, v in item.items() if k != 'timestamp'}
                cleaned_items.append(cleaned_item)

            return cleaned_items
    
    def _start_cleanup_thread(self):
        thread = threading.Thread(target=self._cleanup, daemon=True)
        thread.start()
    
    def _start_cleanup(self):
        def scheduled_cleanup():
            self._cleanup()
            # Schedule next run
            self.scheduler.enter(self.cleanup_interval, 1, scheduled_cleanup)

        # Schedule first run immediately
        self.scheduler.enter(0, 1, scheduled_cleanup)
        threading.Thread(target=self.scheduler.run, daemon=True).start()

    def _cleanup(self):
        current_time = time.time()
        with self.lock:
            original_len = len(self.queue)
            self.queue = deque(
                [item for item in self.queue if current_time - item['timestamp'] <= self.expiry_seconds]
            )
            cleaned = original_len - len(self.queue)
            if cleaned > 0:
                print(f"[MetricQueue] Cleaned {cleaned} expired metrics")

metric_queue = ExpiringMetricQueue(expiry_seconds=60, cleanup_interval=60)


from prometheus_client import Gauge, start_http_server
import psutil
import time
import sys
import os
from flexmetric.config.configuration import CA_PATH, CERT_PATH, KEY_PATH
from flexmetric.logging_module.logger import get_logger
from flexmetric.file_recognition.exec_file import execute_functions
from flexmetric.metric_process.process_commands import process_commands
from flexmetric.metric_process.database_processing import process_database_queries
from flexmetric.metric_process.expiring_queue import metric_queue
import argparse
import os
from flexmetric.metric_process.views import (
    run_flask,
    add_update_metric_route,
    secure_flask_run,
)
import sched
import threading

scheduler = sched.scheduler(time.time, time.sleep)


def arguments():
    parser = argparse.ArgumentParser(
        description="FlexMetric: A flexible Prometheus exporter for commands, databases, scripts, and Python functions."
    )

    # Input type flags
    parser.add_argument(
        "--database",
        action="store_true",
        help="Process database.yaml and queries.yaml to extract metrics from databases.",
    )
    parser.add_argument(
        "--commands",
        action="store_true",
        help="Process commands.yaml to extract metrics from system commands.",
    )
    parser.add_argument(
        "--functions",
        action="store_true",
        help="Process Python functions from the provided path to extract metrics.",
    )
    parser.add_argument(
        "--expose-api",
        action="store_true",
        help="Expose Flask API to receive external metric updates.",
    )

    # Config file paths
    parser.add_argument(
        "--database-config",
        type=str,
        default=None,
        help="Path to the database configuration YAML file.",
    )
    parser.add_argument(
        "--queries-config",
        type=str,
        default=None,
        help="Path to the database queries YAML file.",
    )
    parser.add_argument(
        "--commands-config",
        type=str,
        default=None,
        help="Path to the commands configuration YAML file.",
    )
    parser.add_argument(
        "--functions-dir", type=str, default=None, help="Path to the python files dir."
    )
    parser.add_argument(
        "--functions-file",
        type=str,
        default=None,
        help="Path to the file containing which function to execute",
    )
    parser.add_argument(
        "--port", type=int, default=8000, help="port on which exportor runs"
    )
    parser.add_argument(
        "--host",
        type=str,
        default="0.0.0.0",
        help="The hostname or IP address on which to run the Flask server (default: 0.0.0.0)",
    )
    parser.add_argument(
        "--enable-https",
        action="store_true",
        help="Enable HTTPS for Flask API using SSL certificates.",
    )
    parser.add_argument(
        "--ssl-cert",
        type=str,
        help="Path to the SSL certificate file (cert.pem) for HTTPS.",
    )
    parser.add_argument(
        "--ssl-key",
        type=str,
        help="Path to the SSL private key file (key.pem) for HTTPS.",
    )

    return parser.parse_args()


logger = get_logger(__name__)

logger.info("prometheus is running")


def convert_to_data_type(value):
    if isinstance(value, str) and "%" in value:
        return float(value.strip("%"))
    elif isinstance(value, str) and ("GB" in value or "MB" in value):
        return float(value.split()[0].replace(",", ""))
    return value


gauges = {}


def validate_required_files(mode_name, required_files):
    missing = [desc for desc, path in required_files.items() if path == None]
    if missing:
        logger.error(
            f"Missing {', '.join(missing)} for '{mode_name}' mode. Skipping..."
        )
        return False

    return True


def validate_all_modes(args):
    """
    Validates all selected modes and their required files.

    Args:
        args: Parsed command-line arguments.

    Returns:
        bool: True if at least one valid mode is properly configured, False otherwise.
    """
    has_valid_mode = False

    mode_validations = [
        (
            args.database,
            "database",
            {
                "database-config": args.database_config,
                "queries-config": args.queries_config,
            },
        ),
        (args.commands, "commands", {"commands-config": args.commands_config}),
        (
            args.functions,
            "functions",
            {
                "functions-dir": args.functions_dir,
                "functions-file": args.functions_file,
            },
        ),
        (args.expose_api, "expose-api", {}),
        (args.enable_https,"enable-https",{"ssl-cert":args.ssl_cert,"ssl-key" :args.ssl_key})
    ]

    for is_enabled, mode_name, files in mode_validations:
        if is_enabled:
            if validate_required_files(mode_name, files):
                has_valid_mode = True

    return has_valid_mode


def measure(args):
    exec_result = []
    queue_items = metric_queue.pop_all()
    if len(queue_items) != 0:
        exec_result.extend(queue_items)
    if args.database:
        db_results = process_database_queries(args.queries_config, args.database_config)
        exec_result.extend(db_results)
    if args.functions:
        function_results = execute_functions(args.functions_dir, args.functions_file)
        exec_result.extend(function_results)
    if args.commands:
        cmd_results = process_commands(args.commands_config)
        if cmd_results != None:
            exec_result.extend(cmd_results)
    global gauges

    for data in exec_result:
        results = data["result"]
        labels = data["labels"]
        main_label_value = data.get("main_label", "default_main")
        gauge_name = main_label_value.lower() + "_gauge"

        if gauge_name not in gauges:
            gauge = Gauge(gauge_name, f"{gauge_name} for different metrics", labels)
            gauges[gauge_name] = gauge
        else:
            gauge = gauges[gauge_name]

        for result in results:
            label_values = result["label"]

            if not isinstance(label_values, list):
                # Automatically wrap single label into list for consistency
                label_values = [label_values]

            if len(label_values) != len(labels):
                logger.error(f"Label mismatch: expected {len(labels)} values but got {len(label_values)}")
                continue

            label_dict = dict(zip(labels, label_values))
            # print(label_dict)

            try:
                gauge.labels(**label_dict).set(convert_to_data_type(result["value"]))
            except Exception as ex:
                logger.error(f"Failed to set gauge for labels {label_dict}: {ex}")


def scheduled_measure(args):
    measure(args)
    scheduler.enter(5, 1, scheduled_measure, (args,))


def start_scheduler(args):
    scheduler.enter(0, 1, scheduled_measure, (args,))
    scheduler.run()


def main():
    args = arguments()
    logger.info("Validating configuration...")
    if not validate_all_modes(args):
        logger.error("No valid modes with proper configuration found. Exiting.")
        exit(1)

    logger.info(f"Starting Prometheus metrics server on port {args.port}...")
    logger.info("Starting server")
    # start_http_server(args.port)

    scheduler_thread = threading.Thread(
        target=start_scheduler, args=(args,), daemon=True
    )
    scheduler_thread.start()
    if args.expose_api:
        add_update_metric_route()
    if args.enable_https:
        secure_flask_run(args)
    else:
        run_flask(args.host, args.port)
# # args = arguments()
# # measure(args)
# main()


from setuptools import setup, find_packages

setup(
    name="flexmetric",
    version="0.4.2",
    author="Nikhil Lingadhal",
    description="A secure flexible Prometheus exporter for commands, databases, functions, and scripts.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/nikhillingadhal1999/flexmetric",
    project_urls={
        "Homepage": "https://github.com/nikhillingadhal1999", 
        "Source": "https://github.com/nikhillingadhal1999/flexmetric",
        "Tracker": "https://github.com/nikhillingadhal1999/flexmetric/issues",
    },
    license="MIT",
    packages=find_packages(),
    install_requires=[
        "prometheus_client",
        "PyYAML",
        "psutil",
        "setuptools",
        "wheel",
        "twine",
        "flask"
    ],
    entry_points={
        "console_scripts": [
            "flexmetric = flexmetric.metric_process.prometheus_agent:main"
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



import logging 
import sys


def get_logger(name: str , level = logging.DEBUG) -> logging.Logger:
    logger = logging.getLogger(__name__)
    if not logger.hasHandlers():
        logger.setLevel(level)
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(
                logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
                )
        logger.addHandler(handler)
    return logger




from flask import Flask, request, jsonify, Response
from flexmetric.metric_process.expiring_queue import metric_queue
import argparse
from prometheus_client import generate_latest, REGISTRY, CONTENT_TYPE_LATEST

app = Flask(__name__)


@app.route('/metrics')
def metrics():
    return Response(generate_latest(REGISTRY), mimetype=CONTENT_TYPE_LATEST)


def add_update_metric_route():
    @app.route("/update_metric", methods=["POST"])
    def update_metric():
        try:
            data = request.get_json(force=True)

            # Top-level validation
            if not isinstance(data, dict):
                return jsonify({"status": "invalid structure", "error": "Top-level JSON must be an object"}), 400

            required_keys = {"result", "labels", "main_label"}
            if not required_keys.issubset(data):
                return jsonify({"status": "invalid structure", "error": f"Missing keys: {required_keys - set(data)}"}), 400

            result = data.get("result")
            labels = data.get("labels")
            main_label = data.get("main_label")

            # Type validation
            if not isinstance(result, list) or not all(isinstance(item, dict) for item in result):
                return jsonify({"status": "invalid result", "error": "Result must be a list of dictionaries"}), 400

            if not isinstance(labels, list) or not all(isinstance(label, str) for label in labels):
                return jsonify({"status": "invalid labels", "error": "Labels must be a list of strings"}), 400

            if not isinstance(main_label, str) or not main_label.strip():
                return jsonify({"status": "invalid main_label", "error": "main_label must be a non-empty string"}), 400

            for idx, item in enumerate(result):
                if "label" not in item or "value" not in item:
                    return jsonify({"status": "invalid result item", "error": f"Item {idx} missing 'label' or 'value'"}), 400

                label_values = item["label"]
                value = item["value"]

                if not isinstance(label_values, list) or not all(isinstance(lv, str) for lv in label_values):
                    return jsonify({"status": "invalid label", "error": f"Item {idx} 'label' must be list of strings"}), 400

                if len(label_values) != len(labels):
                    return jsonify({
                        "status": "label count mismatch",
                        "error": f"Item {idx} label count ({len(label_values)}) does not match labels length ({len(labels)})"
                    }), 400

                try:
                    float(value)  # Validate numeric value
                except (ValueError, TypeError):
                    return jsonify({"status": "invalid value", "error": f"Item {idx} value must be numeric"}), 400

            # If all checks pass, queue the metric
            metric_queue.put(data)
            print("Queued:", data)

            return jsonify({"status": "success"}), 200

        except Exception as e:
            return jsonify({"status": "error", "message": str(e)}), 500


def run_flask(host, port):
    app.run(host=host, port=port)
def secure_flask_run(args):
    app.run(host=args.host, port=args.port, ssl_context=(args.ssl_cert, args.ssl_key))